;;; blob handling

;; This is experimental enough not to warrant an egg.
;; If eggified, you must emit an inline file
;; for reasonable performance.

;; Unsafe procedures are marked with a * prefix. The
;; "semisafe" feature must be registered to enable them;
;; otherwise, they are aliases for safe calls.
;; A procedure that makes unsafe calls is also considered
;; unsafe UNLESS it verifies all arguments to the
;; unsafe calls are safe.

;; Strategic use of a few semisafe calls and fixnum-arithmetic
;; can give a speedup comparable to compiling an entire
;; extension unsafe, with a very high safety margin.

(define (make-zero-blob len)
  (let ((b (make-blob len)))
    (blob-zero! b)
    b))
(define (blob-zero! b)
  (let ((v (blob->u8vector/shared b)))
    (do ((i 0 (fx+ i 1))
         (n (u8vector-length v)))
        ((>= i n) b)
      (u8vector-set! v i 0))))
(define (blob-u8-ref b i)                        ; eww!
  (u8vector-ref (blob->u8vector/shared b) i))
(define (blob-u8-set! b i n)                     ; eww!
  (u8vector-set! (blob->u8vector/shared b) i n))
(define (blob-be16-ref b i)                      ; oh god it burns
  (let ((v (blob->u8vector/shared b)))
    (u8vector-be16-ref v i)))
(define (blob-le16-ref b i)
  (let ((v (blob->u8vector/shared b)))
    (u8vector-le16-ref v i)))
(define (blob-le16-set! b i n)
  (let ((v (blob->u8vector/shared b)))
    (u8vector-le16-set! v i n)))
(define (blob-be16-set! b i n)
  (let ((v (blob->u8vector/shared b)))
    (u8vector-set! v (fx+ i 1)
                   (bitwise-and n #xff))
    (u8vector-set! v i
                   (bitwise-and (fxshr n 8) #xff))))

;; move-memory! does bounds checking but does not check negative offsets,
;; which can lead to corruption.  See ticket #142.   Fixed in
;; git 72be4d4 (4.2.14).
(define (blob-substring b i len)
  (when (fx< i 0)
    (error 'blob-substring "illegal negative offset" i))
  (let ((str (make-byte-string len)))
    (move-memory! b str len i 0)
    str))

(define (blob-set-substring! b i str)
  (when (fx< i 0)
    (error 'blob-set-substring! "illegal negative offset" i))
  (move-memory! str b (string-length str) 0 i))

(cond-expand
 ((and compiling semisafe)
  (define (*u8vector-ref v i) (##core#inline "C_u_i_u8vector_ref" v i))
  (define (*u8vector-set! v i n) (##core#inline "C_u_i_u8vector_set" v i n))  
  (define (*byte-string-ref s i) (##core#inline "C_subchar" s i)))
 (else
  (define (*u8vector-ref v i) (u8vector-ref v i))
  (define (*u8vector-set! v i n) (u8vector-set! v i n))
  (define (*byte-string-ref s i) (byte-string-ref s i))))

;; Avoids calling blob->u8vector/shared every reference.
;; Need to manually inline if globally visible -- -local doesn't
;; for some reason, investigate this.
(define (u8vector-le16-ref v i)
  (bitwise-ior (u8vector-ref v i)
               (fxshl (u8vector-ref v (fx+ i 1)) 8)))
(define (*u8vector-le16-ref v i)
  (fxior (*u8vector-ref v i)
         (fxshl (*u8vector-ref v (fx+ i 1)) 8)))
(define (u8vector-le16-set! v i n)
  (u8vector-set! v i
                 (bitwise-and n #xff))
  (u8vector-set! v (fx+ i 1)
                 (bitwise-and (fxshr n 8) #xff)))
(define (*u8vector-le16-set! v i n)
  (*u8vector-set! v i
                 (fxand n #xff))
  (*u8vector-set! v (fx+ i 1)
                 (fxand (fxshr n 8) #xff)))
(define (u8vector-be16-ref v i)
  (bitwise-ior (fxshl (u8vector-ref v i) 8)
               (u8vector-ref v (fx+ i 1))))
(define (*u8vector-be16-ref v i)
  (fxior (fxshl (*u8vector-ref v i) 8)
         (*u8vector-ref v (fx+ i 1))))
(define (u8vector-be16-set! v i n)
  (u8vector-set! v (fx+ i 1)
                 (bitwise-and n #xff))
  (u8vector-set! v i
                 (bitwise-and (fxshr n 8) #xff)))
(define (*u8vector-be16-set! v i n)
  (*u8vector-set! v (fx+ i 1)
                 (fxand n #xff))
  (*u8vector-set! v i
                 (fxand (fxshr n 8) #xff)))

(define (u8vector-substring=? v s voff slen)
  (if (or (< voff 0)
          (> (+ voff slen) (u8vector-length v))
          (> slen (string-length s)))
      (error 'u8vector-substring=? "Illegal (offset . length)" (cons voff slen)) ; *unsafe-only
      (let loop ((vi voff)
                 (si 0))
        (cond ((fx>= si slen)
               #t)
              ((fx= (*u8vector-ref v vi)                      ; Good speedup
                    (char->integer (*byte-string-ref s si)))  ; Little advantage.
               (loop (fx+ vi 1) (fx+ si 1)))
              (else #f)))))

;; Set up network- and host-order access according to this
;; processor architecture.  Note that this is verbose because
;; inlining will not take place when the (define x y) form
;; is used; you must use (define (x a b) (y a b)).

(define (blob-nu16-ref b i) (blob-be16-ref b i))
(define (blob-nu16-set! b i n) (blob-be16-set! b i n))
(define (u8vector-nu16-ref v i) (u8vector-be16-ref v i))
(define (*u8vector-nu16-ref v i) (*u8vector-be16-ref v i))
(define (u8vector-nu16-set! v i n) (u8vector-be16-set! v i n))
(define (*u8vector-nu16-set! v i n) (*u8vector-be16-set! v i n))

(cond-expand
 (little-endian
  (define (blob-hu16-ref b i) (blob-le16-ref b i))
  (define (blob-hu16-set! b i n) (blob-le16-set! b i n))
  (define (u8vector-hu16-ref v i) (u8vector-le16-ref v i))
  (define (*u8vector-hu16-ref v i) (*u8vector-le16-ref v i))
  (define (u8vector-hu16-set! v i n) (u8vector-le16-set! v i n))
  (define (*u8vector-hu16-set! v i n) (*u8vector-le16-set! v i n)))
 (big-endian
  (define (blob-hu16-ref b i) (blob-be16-ref b i))
  (define (blob-hu16-set! b i n) (blob-be16-set! b i n))
  (define (u8vector-hu16-ref v i) (u8vector-be16-ref v i))
  (define (*u8vector-hu16-ref v i) (*u8vector-be16-ref v i))
  (define (u8vector-hu16-set! v i n) (u8vector-be16-set! v i n))
  (define (*u8vector-hu16-set! v i n) (*u8vector-be16-set! v i n))))

